import {
  app,
  BrowserWindow,
  clipboard,
  ipcMain,
  Menu,
  MenuItemConstructorOptions,
  Notification,
  Tray,
  globalShortcut,
} from 'electron'
import * as path from 'path'
import * as isDev from 'electron-is-dev'
import log from 'electron-log'
import installExtension, { REACT_DEVELOPER_TOOLS } from 'electron-devtools-installer'
import * as config from './config.json'
import { KlinkDetails } from '../common/types'
import {
  AuthManager,
  KlinkManager,
  NotificationsManager,
  TargetManager,
} from './services'

import {
  getAuthenticationURL,
  loadTokens,
  getCurrentUser,
  onOauthChange,
  logout,
  refreshTokens,
  shouldRefreshTokens,
} from './services/oauth2-manager'

import {
  createKlink,
  getCurrentUserKlinks,
  readKlink,
  editKlink,
} from './services/api-klink-manager'

import { registerPushy } from './services/api-notifications-manager'

log.info('[App] Starting application')

if (process.platform === 'win32' && isDev) {
  app.setAppUserModelId(process.execPath)
}

// Handle creating/removing shortcuts on Windows when installing/uninstalling.
if (require('electron-squirrel-startup')) {
  app.quit()
}

const closeWindows = () => {
  log.info('[App] Closing existing windows')

  BrowserWindow.getAllWindows().forEach((window) => {
    if (window.isClosable()) {
      window.close()
    }
  })
}

const targetManager = new TargetManager()
const authManager = new AuthManager()
const klinkManager = new KlinkManager(authManager)
const notificationsManager = new NotificationsManager(authManager)

const createWindow = async (): Promise<void> => {
  if (shouldRefreshTokens()) {
    await refreshTokens()
  }

  closeWindows()

  log.info('[App] Creating application window')

  const window = new BrowserWindow({
    width: 800,
    height: 600,
    skipTaskbar: true,
    webPreferences: {
      preload: path.join(__dirname, './preload.js'),
      disableBlinkFeatures: 'Auxclick',
      contextIsolation: true,
      sandbox: true,
    },
    show: false,
  })

  window.on('ready-to-show', () => {
    window.show()
  })

  const url = isDev
    ? 'http://localhost:3000'
    : `file://${path.join(__dirname, '../../build/index.html')}`

  window.loadURL(url)

  if (isDev) {
    window.webContents.openDevTools({
      mode: 'detach',
    })
  }
}

app.on('open-url', async (event, url) => {
  event.preventDefault()

  log.info(`[App] Opening custom protocol URL: ${url}`)

  const { hostname } = new URL(url)

  targetManager.setTarget(`/klink/${hostname}`)

  if (app.isReady()) {
    await createWindow()
  }
})

if (process.platform === 'win32') {
  const url = process.argv.pop()

  if (url && url.startsWith('klink://')) {
    log.info(`[App] Opening custom protocol URL: ${url}`)

    const { hostname } = new URL(url)

    targetManager.setTarget(`/klink/${hostname}`)
  }
}

const gotTheLock = app.requestSingleInstanceLock()
if (!gotTheLock) {
  app.quit()

  log.error('[App] Failed second instance lock')

  process.exit(0)
} else {
  log.info('[App] Got second instance lock')

  let tray: Tray

  const onAppIsReadyNotification = () => {
    new Notification({
      title: 'Klink App',
      body: 'Klink App is ready',
    }).show()
  }

  app
    .whenReady()
    .then(onAppIsReadyNotification)
    .then(async () => {
      log.info('[App] Application ready')

      const shortcutRegistered = globalShortcut.register(
        config.SHORTCUT_CREATE_KLINK,
        () => {
          targetManager.setTarget(null)

          createWindow()
        }
      )

      if (!shortcutRegistered) {
        log.error('[App] Failed registering shortcut')
      }

      if (app.isDefaultProtocolClient('klink')) {
        app.removeAsDefaultProtocolClient('klink')
      }

      if (isDev && process.platform === 'win32') {
        // Set the path of electron.exe and your app.
        // These two additional parameters are only available on windows.
        // Setting this is required to get this working in dev mode.
        app.setAsDefaultProtocolClient('klink', process.execPath, [
          path.resolve(process.argv[1] || ''),
        ])
      } else {
        app.setAsDefaultProtocolClient('klink')
      }

      if (isDev) {
        installExtension(REACT_DEVELOPER_TOOLS)
      }

      if (shouldRefreshTokens()) {
        await refreshTokens()
      }

      const buildContextMenu = () => {
        const authMenuItems: MenuItemConstructorOptions[] = []

        const user = getCurrentUser()
        if (user !== null) {
          const currentUserItem: MenuItemConstructorOptions = {
            label: `Current user: ${user.phoneNumber}`,
          }

          authMenuItems.push(currentUserItem)

          authMenuItems.push({
            label: 'Logout',
            click: () => {
              logout()
              closeWindows()
            },
          })
        } else {
          authMenuItems.push({
            label: 'Login',
            click: () => {
              targetManager.setTarget('/login')

              createWindow()
            },
          })
        }

        return Menu.buildFromTemplate([
          ...authMenuItems,
          { type: 'separator' },
          {
            label: 'Create Klink',
            click: () => {
              targetManager.setTarget(null)

              createWindow()
            },
          },
          {
            label: 'My Klinks',
            click: () => {
              targetManager.setTarget('/my-klinks')

              createWindow()
            },
          },
          {
            label: 'Deklink',
            click: () => {
              targetManager.setTarget('/deklink')

              createWindow()
            },
          },
          {
            label: 'Quit',
            click: () => {
              app.quit()
            },
          },
        ])
      }

      tray = new Tray(path.join(__dirname, '../../assets/icon.png'))
      tray.setToolTip('Klink')
      tray.setContextMenu(buildContextMenu())

      onOauthChange(() => {
        tray.setContextMenu(buildContextMenu())

        log.info('[Tray] Tray menu rebuilt')
      })

      app.on('second-instance', async (e, argv) => {
        log.info('[App] Second instance started')

        if (process.platform !== 'darwin') {
          const url = argv.pop()

          if (url && url.startsWith('klink://')) {
            log.info(`[App] Opening custom protocol URL: ${url}`)

            const { hostname } = new URL(url)

            targetManager.setTarget(`/klink/${hostname}`)

            await createWindow()
          }
        }
      })

      ipcMain.on('is-dev', (event) => {
        event.returnValue = isDev
      })

      ipcMain.on('get-target', (event) => {
        event.returnValue = targetManager.popTarget()
      })

      ipcMain.on('get-clipboard-text', (event) => {
        event.returnValue = clipboard.readText()
      })

      ipcMain.handle('write-clipboard-text', (event, text: string) => {
        clipboard.writeText(text)
      })

      ipcMain.on('get-auth-details', async (event) => {
        event.returnValue = getCurrentUser()
      })

      ipcMain.on('login', async (event, phone: string, password: string) => {
        try {
          const authDetails = await authManager.login(phone, password)

          event.reply('login', authDetails)
        } catch (e) {
          event.reply('login', {
            error: true,
            message: e.message,
          })
        }
      })

      ipcMain.on('confirm-tfa', async (event, code: string) => {
        try {
          await authManager.confirmTfa(code)

          event.reply('confirm-tfa')
        } catch (e) {
          event.reply('confirm-tfa', {
            error: true,
            message: e.message,
          })
        }
      })

      ipcMain.on('create-klink', async (event, klinkRequest: KlinkDetails) => {
        try {
          const token = await createKlink(klinkRequest)

          event.reply('create-klink', token)
        } catch (e) {
          event.reply('create-klink', {
            error: true,
            message: e.message,
          })
        }
      })

      ipcMain.on(
        'edit-klink',
        async (event, token: string, klinkRequest: KlinkDetails) => {
          try {
            await editKlink(token, klinkRequest)

            event.reply('edit-klink')
          } catch (e) {
            event.reply('edit-klink', {
              error: true,
              message: e.message,
            })
          }
        }
      )

      ipcMain.on('read-klink', async (event, token: string) => {
        try {
          const details = await readKlink(token)

          event.reply('read-klink', details)
        } catch (e) {
          event.reply('read-klink', {
            error: true,
            message: e.message,
          })
        }
      })

      ipcMain.on('get-current-user-klinks', async (event, page?: number) => {
        try {
          const response = await getCurrentUserKlinks(page)

          event.reply('get-current-user-klinks', response)
        } catch (e) {
          event.reply('get-current-user-klinks', {
            error: true,
            message: e.message,
          })
        }
      })

      ipcMain.on('generate-klink', async (event, token) => {
        event.returnValue = `${config.API_URL}/t/${token}`
      })

      ipcMain.handle('close-window', () => {
        closeWindows()
      })

      ipcMain.on('login-with-kloke', async (event) => {
        closeWindows()
        try {
          let win = new BrowserWindow({
            width: 800,
            height: 600,
            show: false,
            webPreferences: {
              nodeIntegration: false,
            },
          })

          await win.loadURL(getAuthenticationURL())
          win.show()

          const filter = {
            urls: [`${config.REDIRECT_URI}*`],
          }

          win.webContents.session.webRequest.onBeforeRequest(filter, async ({ url }) => {
            await loadTokens(url)
            win.close()
          })

          win.on('closed', function () {
            win = null
          })
        } catch (e) {
          event.reply('login-with-kloke', {
            error: true,
            message: e.messsage,
          })
        }
      })

      await registerPushy((message) => {
        const notification = new Notification({
          title: 'Message was read',
          body: `"${message.title}" was read by ${message.reader}`,
        })

        notification.on('click', () => {
          targetManager.setTarget(`/klink/${message.token}`)

          createWindow()
        })

        notification.show()
      })

      if (targetManager.hasTarget()) {
        await createWindow()
      }
    })

  //TODO: add oauth2 API to permissions

  // app.on('web-contents-created', (event, contents) => {
  //   log.info(`[App] New window contents created: ${contents.getType()}`)

  //   // Block non-JS navigation https://github.com/doyensec/electronegativity/wiki/LIMIT_NAVIGATION_JS_CHECK
  //   contents.on('will-navigate', (event) => {
  //     event.preventDefault()
  //   })
  //   contents.setWindowOpenHandler(() => {
  //     return { action: 'deny' }
  //   })

  //   // Block all browser APIs https://github.com/doyensec/electronegativity/wiki/PERMISSION_REQUEST_HANDLER_JS_CHECK
  //   contents.session.setPermissionRequestHandler((webContents, permission, callback) => {
  //     return callback(false)
  //   })
  // })

  app.on('window-all-closed', () => {
    log.info('[App] All windows closed')
  })
}
